"""Tests for scanner module."""
from raesl.compile import scanner


def test_whitespace_pat():
    for text in [
        "|",
        " |",
        " |\n",
        "|a",
        " |a",
        "... |",
        "|.... ",
        " ...\n|",
        "|... text",
        "|... ...",
        "#some thing|",
        "#some thing\r|\n",
        " #|",
        "...#\n |text",
        "...#\n ... \n |text",
    ]:
        exp_len = text.find("|")
        text = text[:exp_len] + text[exp_len + 1 :]

        s = scanner.Lexer(None, text, 0, 0, 1, [])
        s.skip_white()
        assert s.offset == exp_len, "Input: '{}'".format(repr(text))


def test_find():
    s = scanner.Lexer(None, " define types\n", 0, 0, 1, [])

    t = s.find("DEFINE_KW")
    assert t, "should find 'define'"
    assert t.tok_text == "define"

    t = s.find("VERB_KW")
    assert not t, "should not find 'verb'"

    t = s.find("TYPE_KW")
    assert t, "should find 'types'"
    assert t.tok_text == "types"

    t = s.find("NL_TK")
    assert t, "should find newline"


def test_not_find_type_kw1():
    s = scanner.Lexer(None, "typess", 0, 0, 1, [])
    t = s.find("TYPE_KW")
    assert not t


def test_not_find_type_kw2():
    s = scanner.Lexer(None, "types-etc", 0, 0, 1, [])
    t = s.find("TYPE_KW")
    assert not t


def test_find_type_kw():
    s = scanner.Lexer(None, "TyPEs", 0, 0, 1, [])
    t = s.find("TYPE_KW")
    assert t
    assert t.tok_text == "TyPEs"


def test_find_name():
    s = scanner.Lexer(None, "label: defines most needs", 0, 0, 1, [])

    t = s.find("NAME")
    assert t, "Should find 'label'"
    assert t.tok_text == "label"

    t = s.find("COLON_TK")
    assert t, "Should find ':'"
    assert t.tok_text == ":"

    t = s.find("NAME")
    assert t, "Should find 'defines'"
    assert t.tok_text == "defines"

    t = s.find("NAME")
    assert t, "Should find 'most'"
    assert t.tok_text == "most"

    t = s.find("NAME")
    assert t, "Should find 'needs'"
    assert t.tok_text == "needs"


def test_name_syntax1():
    s = scanner.Lexer(None, "abc-def", 0, 0, 1, [])

    t = s.find("NAME")
    assert t, "Should find 'abc-def'"
    assert t.tok_text == "abc-def"


def test_name_syntax2():
    s = scanner.Lexer(None, "-def", 0, 0, 1, [])

    t = s.find("NAME")
    assert not t, "Should not find '-def'"

    t = s.find("NONSPACE")
    assert t, "Should find '-def'"
    assert t.tok_text == "-def"


def test_name_syntax3():
    s = scanner.Lexer(None, "abc-", 0, 0, 1, [])

    t = s.find("NAME")
    assert t, "Should find 'abc'"
    assert t.tok_text == "abc"


def test_find_words():
    s = scanner.Lexer(None, "label: defines most needs", 0, 0, 1, [])

    t = s.find("NONSPACE")
    assert t, "Should find 'label:'"
    assert t.tok_text == "label:"

    t = s.find("DEFINE_KW")
    assert not t, "should not find define keyword"

    t = s.find("NONSPACE")
    assert t, "Should find 'defines'"
    assert t.tok_text == "defines"

    t = s.find("NONSPACE")
    assert t, "Should find 'most'"
    assert t.tok_text == "most"

    t = s.find("NONSPACE")
    assert t, "Should find 'needs'"
    assert t.tok_text == "needs"


def test_find_comma():
    s = scanner.Lexer(None, "x,y x, y,", 0, 0, 1, [])

    t = s.find("NONCOMMA")
    assert t, "Should find 'x,y'"
    assert t.tok_text == "x,y"

    t = s.find("NONSPACE")
    assert t, "Should find 'x,'"
    assert t.tok_text == "x,"

    t = s.find("NONCOMMA")
    assert t, "Should find 'y'"
    assert t.tok_text == "y"



"""Excel export utility methods."""

from typing import Any, Dict, Generator, Iterable, List, Optional, Set, Tuple

from openpyxl.worksheet.cell_range import CellRange
from openpyxl.worksheet.table import Table, TableStyleInfo
from openpyxl.worksheet.worksheet import Worksheet
from ragraph.edge import Edge
from ragraph.graph import Graph
from ragraph.node import Node

from raesl.doc import lines
from raesl.excel import text
from raesl.excel.defaults import MONO, WRAP  # noqa


def get_all_tags(nodes: List[Node]) -> List[str]:
    """Get all tagged comment keys from a list of nodes."""
    tags: List[str] = []
    seen: Set[str] = set()
    for node in nodes:
        node_tags = [
            tag for tag in node.annotations.esl_info["tagged_comments"].keys() if tag not in seen
        ]
        tags.extend(node_tags)
        seen.update(node_tags)
    return tags


def make_table(
    ws: Worksheet,
    name: str = "Table",
    min_row: int = 1,
    max_row: int = 1,
    min_col: int = 1,
    max_col: int = 1,
    style: TableStyleInfo = TableStyleInfo(name="TableStyleMedium9", showRowStripes=True),
) -> Table:
    """Make a table of a cell range in a worksheet."""
    ref = CellRange(min_row=min_row, max_row=max_row, min_col=min_col, max_col=max_col)
    table = Table(name=name, displayName=name, ref=ref.coord)
    table.tableStyleInfo = style
    ws.add_table(table)
    return table


def apply_styling(
    ws: Worksheet,
    headers: List[str],
    defaults: Dict[str, Any] = dict(),
    start_row: int = 0,
    end_row: Optional[int] = None,
):
    """Apply styling to columns given some default option dictionary."""
    for i, header in enumerate(headers):
        styles = defaults.get(header, dict()).get("styles", dict())
        for i, row in enumerate(ws.rows):
            if i < start_row:
                continue
            if end_row is not None and i == end_row:
                break
            for cell in row:
                cell.font = styles.get("font", "monospace")

    dims = {}
    for i, row in enumerate(ws.rows):
        if i < start_row:
            continue
        if end_row is not None and i == end_row:
            break
        if i == start_row:
            continue

        for cell in row:
            if cell.value:
                dims[cell.column_letter] = max(
                    (dims.get(cell.column_letter, 0), len(str(cell.value)))
                )
    for col, value in dims.items():
        ws.column_dimensions[col].width = 1.5 * value


def parent_component(requirement: Node, skip: Optional[str] = "world") -> str:
    """Get parent component name of a requirement."""
    path = ".".join(requirement.name.split(".")[:-1])
    return lines.node_path(path, arrows=False, skip=skip) if skip else path


def parent_def(graph: Graph, requirement: Node) -> str:
    """Get the parent (component) definition of a requirement."""
    parent_comp = ".".join(requirement.name.split(".")[:-1])
    parent_def = graph[parent_comp].annotations.esl_info["definition_name"]
    return parent_def


def format_multiline(comments: List[str]) -> str:
    """Format multiline (list) text."""
    return "\n".join(c.rstrip("\\") for c in comments)


def requirement_kind(requirement: Node) -> str:
    """Get requirement kind."""
    info = requirement.annotations.esl_info
    if requirement.kind == "function_spec":
        return info["sub_kind"]
    elif requirement.kind == "design_spec":
        return "design"
    elif requirement.kind == "behavior_spec":
        return "behavior"
    else:
        return requirement.kind


def dedupe(iterable: Iterable) -> List[Any]:
    """Deduplicate any iterable into a list where the first occurrence is preserved."""
    seen: Set[Any] = set()
    unique: List[Any] = list()
    for item in iterable:
        if item in seen:
            continue
        seen.add(item)
        unique.append(item)
    return unique


def yield_functional_dependencies(
    g: Graph, component: Node, flows: Optional[Set[str]]
) -> Generator[Edge, None, None]:
    """Yield all functional dependencies with which a component is involved."""
    if flows is None:
        for e in g.edges_from(component):
            if e.kind == "functional_dependency":
                yield e
    else:
        for e in g.edges_from(component):
            if e.kind == "functional_dependency" and flows.intersection(e.labels):
                yield e


def yield_all_function_specifications(
    g: Graph, functional_dependencies: Iterable[Edge]
) -> Generator[Tuple[Node, Node], None, None]:
    """Yield all function specification nodes and edge combinations
    for a given iterable of dependency edges.
    """
    for e in functional_dependencies:
        yield from yield_function_specifications(g, e)


def yield_function_specifications(
    g: Graph, functional_dependency: Edge
) -> Generator[Tuple[Node, Node], None, None]:
    """Get all function specification nodes corresponding to a functional dependency edge."""
    for spec in get_function_specification_names(functional_dependency):
        yield (functional_dependency, g[spec])


def get_function_specification_names(functional_dependency: Edge) -> List[str]:
    """Get the function specification node names corresponding to a functional dependency edge."""
    return (
        functional_dependency.annotations.get("esl_info", dict())
        .get("reason", dict())
        .get("function_specifications", [])
    )


def yield_active_functions(
    function_specifications: Iterable[Node], active_path: str
) -> Generator[Node, None, None]:
    """Yield all edge function combinations for which the active component matches the given
    path.
    """
    for n in function_specifications:
        if get_goal_active(n[1]) == active_path:
            yield n


def get_goal_active(function_specification: Node) -> Optional[str]:
    """Get the active component in a function specification."""
    esl_info = function_specification.annotations.get("esl_info", dict())
    if esl_info.get("sub_kind", None) != "goal":
        return None
    return esl_info.get("body", dict()).get("active", None)


def yield_passive_functions(
    function_specifications: Iterable[Node], passive_path: str
) -> Generator[Tuple[Node, Node], None, None]:
    """Yield all edge function combinations for which the passive component matches the given
    path.
    """
    for n in function_specifications:
        if get_goal_passive(n[1]) == passive_path:
            yield n


def get_goal_passive(function_specification: Node) -> Optional[str]:
    """Get the receiving component of a goal function."""
    esl_info = function_specification.annotations.get("esl_info", dict())
    if esl_info.get("sub_kind", None) != "goal":
        return None
    return esl_info.get("body", dict()).get("passive", None)


def get_function_flow_variables(
    function_specification: Node, edge: Edge, function_specifications: Set[Node] = set()
) -> Set[str]:
    """Get the variables involved with a functional dependency."""
    esl_info = function_specification.annotations.get("esl_info", dict())

    transformations = set(
        [
            f
            for (e, f) in function_specifications
            if f.annotations["esl_info"]["sub_kind"] == "transformation"
            and f.name in e.annotations["esl_info"]["reason"]["function_specifications"]
            and f.annotations["esl_info"]["body"]["active"] == edge.target.name
        ]
    )

    goal_vars = set(esl_info.get("body", dict()).get("variables", []))

    if transformations:
        trans_vars = set()
        for t in transformations:
            trans_vars = trans_vars.union(
                set(
                    t.annotations.get("esl_info", dict())
                    .get("body", dict())
                    .get("input_variables", [])
                    + t.annotations.get("esl_info", dict())
                    .get("body", dict())
                    .get("output_variables", [])
                )
            )

        flow_vars = goal_vars.intersection(trans_vars)
        return flow_vars if flow_vars else goal_vars
    else:
        return goal_vars


def get_function_subclauses(function_specification: Node) -> List[str]:
    """Get the subclauses belonging to a function specification."""
    esl_info = function_specification.annotations.get("esl_info", dict())
    return esl_info.get("body", dict()).get("subclauses", [])


def get_variable_type(variable: Node) -> str:
    """Get the variable type of a node."""
    return variable.annotations.get("esl_info", dict()).get("type_ref", "")


def build_active_goal_data(
    graph: Graph, edge: Edge, goal: Node, function_specifications: Set[Node] = set()
) -> Generator[Dict[str, Any], None, None]:
    """Create data dictionaries for the outgoing goals in the component overview.

    Keys:
        label: Goal label.
        target: Target component (shared path skipped).
        flows: Sent flow variables (shared path skipped).
        types: Flow variable types.
        subclause: Subclause label, optionally prefixed with numbers for OR-concatenations.
        subject: Subclause subject with prefix skipped.
        comparison: Abbreviated comparison (<, <=, ==, >=, >, ++, --).
        bound: Comparison variable or value.
        unit: Comparison value unit.
    """
    label = goal.name
    active = get_goal_active(goal)
    target = edge.target.name

    flows = get_function_flow_variables(goal, edge, function_specifications)
    types = set([get_variable_type(graph[f]) for f in flows if graph.node_dict.get(f, False)])
    types = ", ".join(types)
    subclauses = get_function_subclauses(goal)

    # Skip everything in common, except the last part, the flow's own name.
    active_skip = active[: active.rfind(".") + 1] if active else ""
    stripped_flows = [text.strip_prefix(f, active_skip) for f in flows]
    display_flows = ", ".join(stripped_flows)

    common_parts = text.get_common_parts(flows)
    skip_prefix = ".".join(common_parts[:-1])

    label = text.strip_prefix(label, active_skip)
    target = text.strip_prefix(target, active_skip)

    if not subclauses:
        data = dict(label=label, target=target, flows=display_flows, types=types)
        yield data

    for sc in subclauses:
        for sc_data in build_subclause_data(sc, skip_prefix):
            data = dict(label=label, target=target, flows=display_flows, types=types, **sc_data)
            yield data


def build_passive_goal_data(
    graph: Graph, edge: Edge, goal: Node, function_specifications: Set[Node]
) -> Generator[Dict[str, Any], None, None]:
    """Create data dictionaries for the incoming goals in the component overview.

    Keys:
        label: Goal label.
        target: Target component (shared path skipped).
        flows: Sent flow variables (shared path skipped).
        types: Flow variable types.
        subclause: Subclause label, optionally prefixed with numbers for OR-concatenations.
        subject: Subclause subject with prefix skipped.
        comparison: Abbreviated comparison (<, <=, ==, >=, >, ++, --).
        bound: Comparison variable or value.
        unit: Comparison value unit.
    """
    label = goal.name
    active = edge.target.name
    passive = get_goal_passive(goal)
    flows = get_function_flow_variables(goal, passive, function_specifications)
    types = set([get_variable_type(graph[f]) for f in flows if graph.node_dict.get(f, False)])
    types = ", ".join(types)
    subclauses = get_function_subclauses(goal)

    # Skip everything in common, except the last part, the flow's own name.
    passive_skip = passive[: passive.rfind(".") + 1] if passive else ""
    stripped_flows = [text.strip_prefix(f, passive_skip) for f in flows]
    display_flows = ", ".join(stripped_flows)

    common_parts = text.get_common_parts(flows)
    skip_prefix = ".".join(common_parts[:-1])

    label = text.strip_prefix(label, passive_skip)
    active = text.strip_prefix(active, passive_skip)

    if not subclauses:
        data = dict(label=label, source=active, flows=display_flows, types=types)
        yield data

    for sc in subclauses:
        for sc_data in build_subclause_data(sc, skip_prefix):
            data = dict(label=label, source=active, flows=display_flows, types=types, **sc_data)
            yield data


def build_subclause_data(
    subclause_info: Dict[str, Any], skip_prefix: str
) -> Generator[Dict[str, Any], None, None]:
    """Create a subclause data dictionary for each OR clause for use in the component overview.

    Keys:
        subclause: Subclause label, optionally prefixed with numbers for OR-concatenations.
        subject: Subclause subject with prefix skipped.
        comparison: Abbreviated comparison (<, <=, ==, >=, >, ++, --).
        bound: Comparison variable or value.
        unit: Comparison value unit.
    """
    name = subclause_info.get("name", None)
    if name is None:
        return  # No label, no joy.
    bodies = subclause_info.get("body", [])  # concatenated OR clauses
    for i, body in enumerate(bodies):
        if len(bodies) == 1:
            subclause = name
        else:
            subclause = f"OR-{i}-{name}"
        data = build_subclause_body_data(body, skip_prefix)
        data["subclause"] = subclause
        yield data


def build_subclause_body_data(subclause_body: Dict[str, Any], skip_prefix: str) -> Dict[str, Any]:
    """Create a subclause's body data dict for use in the component overview.

    Keys:
        subject: Subclause subject with prefix skipped.
        comparison: Abbreviated comparison (<, <=, ==, >=, >, ++, --).
        bound: Comparison variable or value.
        unit: Comparison value unit.
    """
    result = dict()

    subject = subclause_body.get("subject", "")
    result["subject"] = text.strip_prefix(subject, skip_prefix)

    comparison = subclause_body.get("comparison", None)
    if comparison is not None:
        result["comparison"] = text.abbreviate_comparison(comparison)

    bound = subclause_body.get("bound", None)
    if isinstance(bound, str):  # Other variable
        result["bound"] = text.strip_prefix(bound, skip_prefix)
    elif isinstance(bound, dict):  # value / unit
        result["bound"] = bound.get("value", None)
        result["unit"] = bound.get("unit", None).strip("[]")

    return result



from ragraph.graph import Graph

import raesl.plot


def test_hierarchy_diagram(pump_example_graph: Graph, check_digraph):
    graph = pump_example_graph

    for depth in [1, 2, 3]:
        fname = f"hierarchy_diagram_{depth}.txt"
        digraph = raesl.plot.hierarchy_diagram(
            graph,
            roots=[graph["world"]],
            levels=depth,
        )
        check_digraph(digraph, fname)


def test_functional_context_diagram(pump_example_graph: Graph, check_digraph):
    """Test functional context diagram."""
    graph = pump_example_graph

    styles = [
        raesl.plot.Style(diagram=dict(show_root_children=False, show_neighbour_children=False)),
        raesl.plot.Style(diagram=dict(show_root_children=True, show_neighbour_children=False)),
        raesl.plot.Style(diagram=dict(show_root_children=False, show_neighbour_children=True)),
        raesl.plot.Style(diagram=dict(show_root_children=True, show_neighbour_children=True)),
    ]

    for idx, style in enumerate(styles):
        fname = f"functional_context_diagram_{idx}.txt"
        digraph = raesl.plot.functional_context_diagram(
            graph,
            root=graph["world.drive-mechanism.motor"],
            degree=1,
            style=style,
        )
        check_digraph(digraph, fname)


def test_functional_traceability_diagram(pump_example_graph: Graph, check_digraph):
    graph = pump_example_graph

    styles = [
        raesl.plot.Style(diagram=dict(show_function_dependencies=False)),
        raesl.plot.Style(diagram=dict(show_function_dependencies=True)),
    ]

    for idx, style in enumerate(styles):
        fname = f"function_chain_diagram_pump_{idx}.txt"
        digraph = raesl.plot.function_traceability_diagram(
            graph,
            root=graph["world.drive-mechanism.convert-power-potential"],
            levels=3,
            style=style,
        )
        check_digraph(digraph, fname)


def test_functional_dependency_diagram(pump_example_graph: Graph, check_digraph):
    graph = pump_example_graph
    depths = [1, 2, 3, 4]
    for depth in depths:
        fname = f"functional_dependency_diagram_{depth}.txt"
        digraph = raesl.plot.functional_dependency_diagram(
            graph,
            root=graph["world"],
            levels=depth,
        )
        check_digraph(digraph, fname)


def test_function_chain_diagram(pump_example_graph: Graph, check_digraph):
    graph = pump_example_graph
    start_points = [
        (
            graph["world.drive-mechanism.power-source"],
            [graph["world.drive-mechanism.power-source.convert-potential"]],
        )
    ]
    end_points = [(graph["world.pump"], [graph["world.pump.convert-torque"]])]
    style = raesl.plot.Style(diagram=dict(show_variables=True))

    for levels in [1, 2]:
        fname = f"function_chain_diagram_{levels}.txt"
        digraph = raesl.plot.function_chain_diagram(
            graph,
            start_points=start_points,
            end_points=end_points,
            levels=levels,
            style=style,
        )
        check_digraph(digraph, fname)



import difflib
import json
import shutil
import uuid
from pathlib import Path

import pytest

TESTS = Path(__file__).parent
UPDATE_PATH = TESTS / ".update"
UPDATE = UPDATE_PATH.exists()
INSPECT_PATH = TESTS / ".inspect"
INSPECT = INSPECT_PATH.exists()
SPECS_PATH = TESTS / "data" / "specs"


# Fix UUID generation during tests:
def generate_int_uuids():
    index = 0
    while True:
        yield uuid.UUID(int=index)
        index += 1


@pytest.fixture(autouse=True, scope="function")
def reset_uuids():
    int_uuid = generate_int_uuids()
    uuid.uuid4 = lambda: next(int_uuid)


@pytest.fixture
def update():
    return UPDATE


@pytest.fixture
def inspect():
    return INSPECT


@pytest.fixture
def datadir():
    return (TESTS / "data").relative_to(Path.cwd())


@pytest.fixture
def gendir():
    path = (TESTS / "generated").relative_to(Path.cwd())
    path.mkdir(exist_ok=True)
    return path


@pytest.fixture
def docsdir() -> Path:
    return TESTS.parent / "docs"


@pytest.fixture
def docs_gendir(docsdir: Path) -> Path:
    path = docsdir / "generated"
    path.mkdir(exist_ok=True)
    return path


@pytest.fixture(params=SPECS_PATH.glob("*.esl"))
def spec(request) -> Path:
    return request.param.relative_to(Path.cwd())


@pytest.fixture(params=(TESTS / "data" / "good_examples").glob("*.esl"))
def good_example(request):
    return request.param.relative_to(Path.cwd())


@pytest.fixture(params=(TESTS / "data" / "bad_examples").glob("*.esl"))
def bad_example(request):
    return request.param.relative_to(Path.cwd())


@pytest.fixture
def pump_example_graph(datadir: Path):
    from raesl.compile import to_graph

    fpath = datadir / "specs" / "pump_example.esl"
    graph = to_graph(fpath)
    return graph


@pytest.fixture
def verify_content():
    def verify(content: str, reference: Path, start=0, stop=None):
        if UPDATE:
            reference.write_text(content)

        try:
            assert (
                content.splitlines()[start:stop] == reference.read_text().splitlines()[start:stop]
            ), "Content lines should match."
        except AssertionError as e:
            if reference.suffix == ".yaml" and "*id" in e.args[0]:
                pass  # Pass on finicky YAML ID errors.
            else:
                raise e

    return verify


@pytest.fixture
def check_diff():
    """Check whether there is a diff w.r.t. to the reference path."""

    def _check_diff(path: Path, ref_path: Path, sort: bool = False):
        text = path.read_text(encoding="utf-8") if path.exists() else ""
        ref_text = ref_path.read_text(encoding="utf-8") if path.exists() else ""

        lines = text.splitlines(False)
        ref_lines = ref_text.splitlines(False)
        if sort:
            lines, ref_lines = sorted(lines), sorted(ref_lines)

        diff = difflib.unified_diff(
            ref_lines,
            lines,
            fromfile=str(ref_path),
            tofile=str(path),
            lineterm="",
        )
        diffstr = "\n".join(diff)

        if diffstr and UPDATE:
            if not ref_path.parent.exists():
                ref_path.parent.mkdir(parents=True, exist_ok=True)
            shutil.copy2(path, ref_path)
        else:
            assert not diffstr, diffstr

    return _check_diff


@pytest.fixture
def check_digraph(tmpdir, datadir, check_diff, gendir, inspect):
    """Digraph output check function."""
    import graphviz

    def _check_digraph(digraph: graphviz.Digraph, fname: Path):
        digraph.comment = "TEST"
        if inspect:
            digraph.view(directory=gendir, filename=fname)
        dot_path = Path(digraph.save(directory=tmpdir, filename=fname))
        ref_path = Path(datadir / "dot" / fname)
        check_diff(dot_path, ref_path, sort=True)

    return _check_digraph


@pytest.fixture
def check_plotly(tmpdir, datadir, check_diff, inspect):
    """Plotly output check function."""
    from plotly.graph_objs import Figure

    def _check_plotly(fig: Figure, fname: str):
        """Checking if the figure data and shapes are equal to the data and shapes
        stored in the reference file.

        Arguments:
            fig: The figure to be tested.
            file_path: The relative path to the reference file.
            update: Whether to update the reference file.
        """
        if inspect:
            fig.show()

        plotly_json = fig.to_plotly_json()

        plotly_json["layout"]["shapes"] = sorted(
            plotly_json["layout"]["shapes"], key=lambda x: str(x)
        )

        for idx, data in enumerate(plotly_json["data"]):
            for key in ["x", "y", "text"]:
                if plotly_json["data"][idx].get("text", None) is None:
                    # Text data is not always present at every plot element.
                    continue
                plotly_json["data"][idx][key] = sorted(data[key], key=lambda x: x)

        (tmpdir / fname).write_text(
            json.dumps(plotly_json, sort_keys=True, indent=2), encoding="utf-8"
        )

        check_diff(tmpdir / fname, datadir / "plotly" / fname, sort=True)

    return _check_plotly



"""Not really tests, but sanity checks for required properties of the state machines.
(Mostly the line matcher machines, but also a few for the high level ESL_MACHINE.)
"""
from typing import TYPE_CHECKING

from raesl.compile import esl_lines, scanner

if TYPE_CHECKING:
    from raesl.compile.state_machine import Location


def collect_locations(loc: "Location"):
    """Collect locations of a state machine.

    Arguments:
        loc: First location.

    Returns:
        All reachable locations from 'loc'.
    """
    seen_locs = set([loc])
    notdone_locs = [loc]
    while notdone_locs:
        loc = notdone_locs.pop()
        for edge in loc.out_edges:
            if edge.dest not in seen_locs:
                seen_locs.add(edge.dest)
                notdone_locs.append(edge.dest)
    return seen_locs


def collect_tok_types(loc: "Location"):
    """Collect used token types of locations reachable from the given start location."""
    tok_types = set()
    for loc in collect_locations(loc):
        for edge in loc.out_edges:
            tok_types.add(edge.tok_type)

    return tok_types


def test_tokens_exist():
    """Check that tokens used in the line state machines actually exist."""
    TOKEN_NAMES = set(scanner.TOKENS.keys()) | set(["NL_TK", "EOF_TK", "epsilon"])

    bad = False
    for machine in esl_lines.get_all_line_machines():
        tok_types = collect_tok_types(machine.initial_loc)
        for tok_type in tok_types:
            if tok_type not in TOKEN_NAMES:
                msg = "Token {} is used in machine {}, but not defined in the scanner."
                print(msg.format(tok_type, machine.name))
                bad = True

    assert not bad


def test_tokens_used():
    """Check that existing tokens in the scanner are actually used."""
    available = set(scanner.TOKENS.keys()) | set(["NL_TK"])
    # 'EOF_TK' and 'epsilon' are both not explicitly used in line matchers.

    for machine in esl_lines.get_all_line_machines():
        tok_types = collect_tok_types(machine.initial_loc)
        available.difference_update(tok_types)

    assert not available


def test_final_accept():
    """Check that an accepting state is also the final state, no further tokens are
    allowed.
    """
    bad = False
    for machine in esl_lines.get_all_line_machines():
        for loc in collect_locations(machine.initial_loc):
            if loc.accept:
                if loc.out_edges:
                    msg = "Accepting location {} in machine {} has outgoing edges."
                    print(msg.format(loc.name, machine.name))
                    bad = True

    assert not bad


def test_accept_final():
    """Check that a final state (no outgoing edges) is also an accepting state."""
    bad = False
    for machine in esl_lines.get_all_line_machines():
        for loc in collect_locations(machine.initial_loc):
            if not loc.out_edges:
                if not loc.accept:
                    msg = (
                        "Deadlock location {} in machine {}: not accepting and "
                        "no outgoing edges."
                    )
                    print(msg.format(loc.name, machine.name))
                    bad = True

    assert not bad


def test_dfa():
    """Check that the line machines are deterministic with respect to tokens."""
    bad = False
    for machine in esl_lines.get_all_line_machines():
        for loc in collect_locations(machine.initial_loc):
            tok_types = set(edge.tok_type for edge in loc.out_edges)
            if len(tok_types) != len(loc.out_edges):
                msg = "Location {} in machine {} is not deterministic."
                print(msg.format(loc.name, machine.name))
                bad = True

    assert not bad


def test_esl_lang_machine():
    """Check that the language state machine only uses tokens representing line
    machines or 'epsilon'.
    """
    bad = False
    line_machine_names = esl_lines.get_line_machine_names()
    for loc in collect_locations(esl_lines.ESL_MACHINE.initial_loc):
        for edge in loc.out_edges:
            if edge.tok_type == "epsilon":
                continue
            if edge.tok_type in line_machine_names:
                continue

            msg = "Location {} in the ESL language machine uses invalid token {}."
            print(msg.format(loc.name, edge.tok_type))
            bad = True

    assert not bad


def test_esl_lang_no_deadlock():
    """Check that the final state (no outgoing edges) is an accepting state."""
    bad = False
    machine = esl_lines.ESL_MACHINE
    for loc in collect_locations(machine.initial_loc):
        if not loc.out_edges:
            if not loc.accept:
                msg = (
                    "Deadlock location '{}' in machine '{}': not accepting and "
                    "no outgoing edges."
                )
                print(msg.format(loc.name, machine.name))
                bad = True

    assert not bad



"""Tests for the view_funcs module."""

import pytest

from raesl.plot import view_funcs

funcs = [
    view_funcs.multi_domain,
    view_funcs.hierarchy,
    view_funcs.function_chain,
    view_funcs.functional_context,
    view_funcs.functional_dependency,
    view_funcs.traceability,
]


@pytest.fixture(params=funcs)
def view_func(request):
    """Returns each view_func once."""
    return request.param


@pytest.fixture
def view_func_kwargs_graph(view_func, pump_example_graph):
    """Creates a combination of view_func, kwargs and graph data."""
    graph = pump_example_graph

    if view_func == view_funcs.hierarchy:
        kwargs = dict(roots=graph.roots, levels=4)

    elif view_func == view_funcs.function_chain:
        kwargs = dict(
            start_points=[
                (
                    graph["world.drive-mechanism.power-source"],
                    [graph["world.drive-mechanism.power-source.convert-potential"]],
                )
            ],
            end_points=[(graph["world.pump"], [graph["world.pump.convert-torque"]])],
        )

    elif view_func == view_funcs.functional_context:
        kwargs = dict(root=graph["world.drive-mechanism"], degree=2)

    elif view_func == view_funcs.functional_dependency:
        kwargs = dict(root=graph["world"], levels=4)

    elif view_func == view_funcs.traceability:
        kwargs = dict(root=graph["world.drive-mechanism.convert-power-potential"], levels=2)

    else:
        kwargs = dict()

    return view_func, kwargs, graph


def test_preserve_graph(view_func_kwargs_graph):
    """Check whether original Graph data remains untouched by view_funcs."""
    view_func, kwargs, graph = view_func_kwargs_graph
    ref = graph.json_dict
    view_func(graph, **kwargs)
    assert graph.json_dict == ref, "Graph data should be untouched by the view_func."



"""Module to export as an Excel workbook."""

from pathlib import Path
from typing import Dict, List, Optional, Union

from openpyxl.workbook import Workbook
from ragraph.graph import Graph

from raesl import logger
from raesl.doc.locales import register_locale
from raesl.excel import defaults, sheets
from raesl.utils import get_scoped_nodes

# Compile module is excluded during docs generation.
try:
    from raesl.compile import to_graph
except ImportError:
    pass


def convert(
    *paths: Union[str, Path],
    output: Optional[Union[str, Path]] = defaults.OUTPUT,
    scopes: Dict[str, Optional[int]] = defaults.SCOPES,
    language: str = "en",
) -> Workbook:
    """Write (part of) an ESL specification to a Excel workbook.

    Arguments:
        paths: ESL specification paths.
        output: Optional Workbook output path (will be overwritten without warning).
        scopes: Dictionary of component paths to relative depths of subcomponents to
            include as scopes for the generated output. Defaults to the complete tree.
        language: Output language.

    Returns:
        Excel workbook instance.
    """

    graph = to_graph(*paths)
    return overview(graph, output, scopes, language=language)


def overview(
    graph: Graph,
    output: Optional[Union[str, Path]] = defaults.OUTPUT,
    scopes: Dict[str, Optional[int]] = defaults.SCOPES,
    language: str = "en",
) -> Workbook:
    """Write (part of) an ESL specification to a Excel workbook.

    Arguments:
        graph: Compiled ESL graph.
        output: Optional Workbook output path (will be overwritten without warning).
        scopes: Dictionary of component paths to relative depths of subcomponents to
            include as scopes for the generated output. Defaults to the complete tree.
        language: Output language.

    Returns:
        Excel workbook instance.
    """
    # Create workbook, but delete default sheet.
    register_locale(language)

    wb = Workbook()
    wb.remove(wb["Sheet"])

    # Derive components from scopes.
    components = [node for node in get_scoped_nodes(graph, scopes) if node.kind == "component"]
    if not components:
        raise ValueError(f"No components found in selected scopes ('{scopes}'). Please reconsider.")

    # Add sheets.
    _, components = sheets.add_components_sheet(wb, components)
    _, goals = sheets.add_goals_sheet(wb, graph, components)
    _, transformations = sheets.add_transformations_sheet(wb, graph, components)
    _, designs = sheets.add_designs_sheet(wb, graph, components)
    _, behaviors = sheets.add_behaviors_sheet(wb, graph, components)
    _, needs = sheets.add_needs_sheet(wb, graph, components)
    _, variables = sheets.add_variable_sheet(wb, graph, components)

    sheets.add_overview_sheet(
        wb, graph, components, goals, transformations, designs, behaviors, needs
    )

    # Protect all sheets.
    for sheet in wb.sheetnames:
        wb[sheet].protection = defaults.SHEETPROTECTION

    # Write output if path is given.
    if output:
        wb.save(str(output))
    return wb


def component_overview(
    graph: Graph,
    component_path: str,
    flow_labels: Optional[List[str]] = None,
    output: Optional[Union[str, Path]] = None,
    language: str = "en",
) -> Workbook:
    """Create a component overview Excel sheet. Somewhat like a free body diagram for a single
    component. Meant to be more discrete but also compact than the complete overview Excel.

    Arguments:
        graph: Compiled ESL graph.
        component_path: ESL specification paths.
        flow_labels: Flow types to include in sheets.
        output: Optional Workbook output path (will be overwritten without warning).
        language: Output language.

    Returns:
        Excel workbook instance.
    """
    register_locale(language)

    # Create workbook, but delete default sheet.
    wb = Workbook()
    wb.remove(wb["Sheet"])

    try:
        if not component_path.startswith("world."):
            component_path = "world." + component_path
        component = graph[component_path]
    except KeyError:
        logger.error("Could not find component!")
        return

    sheets.add_component_active_goals_sheet(wb, graph, component, flow_labels)
    sheets.add_component_passive_goals_sheet(wb, graph, component, flow_labels)

    if output:
        wb.save(str(output))
    return wb



"""Tests for the typechecking utils module."""
from raesl.compile.typechecking import utils


def test_nomultiple():
    arguments = [0, 1, 2, 3, 4, 5]
    result = utils.split_arguments(len(arguments), None, arguments)
    assert result == [[a] for a in arguments]


def test_multiple1():
    arguments = [0, 1, 2, 3, 4, 5]
    params_length = 5

    result = utils.split_arguments(params_length, 0, arguments)
    assert result == [[0, 1], [2], [3], [4], [5]]

    result = utils.split_arguments(params_length, 1, arguments)
    assert result == [[0], [1, 2], [3], [4], [5]]

    result = utils.split_arguments(params_length, 2, arguments)
    assert result == [[0], [1], [2, 3], [4], [5]]

    result = utils.split_arguments(params_length, 3, arguments)
    assert result == [[0], [1], [2], [3, 4], [5]]

    result = utils.split_arguments(params_length, 4, arguments)
    assert result == [[0], [1], [2], [3], [4, 5]]


def test_multiple2():
    arguments = [0, 1, 2, 3, 4]
    params_length = 5

    # 'multiple' is 1 argument
    result = utils.split_arguments(params_length, 0, arguments)
    assert result == [[a] for a in arguments]

    result = utils.split_arguments(params_length, 1, arguments)
    assert result == [[a] for a in arguments]

    result = utils.split_arguments(params_length, 2, arguments)
    assert result == [[a] for a in arguments]

    result = utils.split_arguments(params_length, 3, arguments)
    assert result == [[a] for a in arguments]

    result = utils.split_arguments(params_length, 4, arguments)
    assert result == [[a] for a in arguments]


def test_multiple3():
    arguments = ["a", "b", "c"]
    params_length = 1

    # 1 multi-value parameter.
    result = utils.split_arguments(params_length, 0, arguments)
    assert result == [arguments]



"""Excel sheet generation."""

from typing import List, Optional, Tuple

from openpyxl.workbook import Workbook
from openpyxl.worksheet.worksheet import Worksheet
from ragraph.graph import Graph
from ragraph.node import Node

from raesl.doc import utils as doc_utils
from raesl.doc.lines import node_path, var_path
from raesl.doc.sections import get_var_table_row_elements
from raesl.excel import text, utils
from raesl.excel.defaults import OPTIONS


def add_components_sheet(wb: Workbook, components: List[Node]) -> Tuple[Worksheet, List[Node]]:
    """Add a components overview sheet to an Excel workbook.

    Arguments:
        wb: Excel workbook to add the components sheet to.
        components: List of component nodes.

    Returns:
        Components worksheet instance.
    """
    ws = wb.create_sheet("Components")

    # Handle headers.
    tags = utils.get_all_tags(components)
    headers = [
        "instance path",
        "component definition",
        "parent component",
        "comments",
    ] + tags
    ws.append(headers)

    # Handle content.
    def write_component_row(ws: Worksheet, tags: List[str], component: Node):
        """Write a component sheet row."""
        info = component.annotations.esl_info
        row = [
            node_path(component.name, italic=False, arrows=False, skip="world"),
            info["definition_name"],
            (
                node_path(component.parent.name, italic=False, arrows=False, skip="world")
                if component.parent
                else None
            ),
            utils.format_multiline(info["comments"]),
        ]
        for tag in tags:
            comments = info["tagged_comments"].get(tag, [])
            row.append(utils.format_multiline(comments))
        ws.append(row)

    for component in components:
        write_component_row(ws, tags, component)

    # Handle styling.
    utils.apply_styling(ws, headers, defaults=OPTIONS)
    utils.make_table(
        ws,
        name="components",
        min_row=1,
        max_row=len(components) + 1,
        min_col=1,
        max_col=len(headers),
    )
    return ws, components


def add_goals_sheet(
    wb: Workbook, graph: Graph, components: List[Node]
) -> Tuple[Worksheet, List[Node]]:
    """Add a goal requirements sheet to an Excel workbook.

    Arguments:
        wb: Excel workbook to add the goals sheet to.
        graph: Graph to fetch goals from.
        components: Component nodes to fetch goals for.

    Returns:
        Goal requirements worksheet instance.
    """
    ws = wb.create_sheet("Goals")

    # Get requirements.
    requirements = utils.dedupe(
        [
            r
            for c in components
            for r in doc_utils.get_component_goals(c, graph, constraint=False, inherited=False)
        ]
        + [
            r
            for c in components
            for r in doc_utils.get_component_goals(c, graph, constraint=False, inherited=True)
        ]
        + [
            r
            for c in components
            for r in doc_utils.get_component_goals(c, graph, constraint=True, inherited=False)
        ]
        + [
            r
            for c in components
            for r in doc_utils.get_component_goals(c, graph, constraint=True, inherited=True)
        ]
    )

    # Handle headers.
    default_headers = [
        "instance path",
        "component definition",
        "form",
        "source component",
        "auxiliary",
        "verb",
        "variables",
        "preposition",
        "target component",
        "subclauses",
        "comments",
    ]
    tags = utils.get_all_tags(requirements)
    headers = default_headers + tags
    ws.append(headers)

    # Handle content.
    for requirement in requirements:
        info = requirement.annotations.esl_info
        body = info["body"]

        default_content = [
            node_path(requirement.name, italic=False, arrows=False, skip="world"),
            utils.parent_def(graph, requirement),
            info["form"],
            node_path(body["active"], italic=False, arrows=False, skip="world"),
            body["auxiliary"],
            body["verb"],
            ", ".join(
                var_path(
                    graph[var],
                    italic=False,
                    arrows=False,
                    skip="world",
                )
                for var in body["variables"]
            ),
            body["preposition"],
            node_path(body["passive"], italic=False, arrows=False, skip="world"),
            text.subclauses_text(requirement, graph, skip="world", spaces=0),
            utils.format_multiline(info["comments"]),
        ]
        tagged_content = [
            utils.format_multiline(info["tagged_comments"].get(tag, [])) for tag in tags
        ]
        ws.append(default_content + tagged_content)

    # Handle styling.
    utils.make_table(
        ws,
        "goals",
        min_row=1,
        max_row=1 + len(requirements),
        min_col=1,
        max_col=len(headers),
    )
    utils.apply_styling(ws, headers, defaults=OPTIONS)
    return ws, requirements


def add_transformations_sheet(
    wb: Workbook, graph: Graph, components: List[Node]
) -> Tuple[Worksheet, List[Node]]:
    """Add a transformation requirements sheet to an Excel workbook.

    Arguments:
        wb: Excel workbook to add the transformations sheet to.
        graph: Graph to fetch transformation nodes from.
        components: Component nodes to fetch transformations for.

    Returns:
        Transformation requirements worksheet instance.
    """
    ws = wb.create_sheet("Transformations")

    # Select requirements.
    requirements = utils.dedupe(
        [
            r
            for c in components
            for r in doc_utils.get_component_transformations(c, graph, constraint=False)
        ]
        + [
            r
            for c in components
            for r in doc_utils.get_component_transformations(c, graph, constraint=True)
        ]
    )

    # Handle headers.
    default_headers = [
        "instance path",
        "component definition",
        "form",
        "auxiliary",
        "verb",
        "input_variables",
        "preposition",
        "output_variables",
        "subclauses",
        "comments",
    ]
    tags = utils.get_all_tags(requirements)
    headers = default_headers + tags
    ws.append(headers)

    for requirement in requirements:
        info = requirement.annotations.esl_info
        body = info["body"]

        default_content = [
            node_path(requirement.name, italic=False, arrows=False, skip="world"),
            utils.parent_def(graph, requirement),
            info["form"],
            body["auxiliary"],
            body["verb"],
            ", ".join(
                var_path(graph[var], italic=False, arrows=False, skip="world")
                for var in body["input_variables"]
            ),
            body["preposition"],
            ", ".join(
                var_path(graph[var], italic=False, arrows=False, skip="world")
                for var in body["output_variables"]
            ),
            text.subclauses_text(requirement, graph, skip="world", spaces=0),
            utils.format_multiline(info["comments"]),
        ]
        tagged_content = [
            utils.format_multiline(info["tagged_comments"].get(tag, [])) for tag in tags
        ]
        ws.append(default_content + tagged_content)

    # Handle styling.
    utils.make_table(
        ws,
        "tranformations",
        min_row=1,
        max_row=1 + len(requirements),
        min_col=1,
        max_col=len(headers),
    )
    utils.apply_styling(ws, headers, defaults=OPTIONS)

    return ws, requirements


def add_designs_sheet(
    wb: Workbook, graph: Graph, components: List[Node]
) -> Tuple[Worksheet, List[Node]]:
    """Add a design requirements sheet to an Excel workbook.

    Arguments:
        wb: Excel workbook to add the design requirements sheet to.
        graph: Graph to fetch designrule nodes from.
        components: Component nodes to fetch design requirements for.

    Returns:
        Design requirements worksheet instance.
    """
    ws = wb.create_sheet("Design requirements")

    # Select requirements.
    requirements = utils.dedupe(
        [r for c in components for r in doc_utils.get_component_designs(c, graph, constraint=False)]
        + [
            r
            for c in components
            for r in doc_utils.get_component_designs(c, graph, constraint=True)
        ]
    )

    # Handle headers.
    default_headers = [
        "instance path",
        "component definition",
        "form",
        "subject",
        "auxiliary",
        "comparison",
        "bound",
        "subclauses",
        "comments",
    ]
    tags = utils.get_all_tags(requirements)
    headers = default_headers + tags
    ws.append(headers)
    rows = 1

    # Handle content.
    for requirement in requirements:
        info = requirement.annotations.esl_info
        for body in info["body"]:
            default_content = [
                node_path(requirement.name, italic=False, arrows=False, skip="world"),
                utils.parent_def(graph, requirement),
                info["form"],
                var_path(graph[body["subject"]], italic=False, arrows=False, skip="world"),
                "{}{}".format("EITHER " if len(info["body"]) > 1 else "", body["auxiliary"]),
                body["comparison"],
                "{} {}".format(body["bound"]["value"], body["bound"]["unit"]),
                text.subclauses_text(requirement, graph, skip="world", spaces=0),
                utils.format_multiline(info["comments"]),
            ]
            tagged_content = [
                utils.format_multiline(info["tagged_comments"].get(tag, [])) for tag in tags
            ]
            ws.append(default_content + tagged_content)
            rows += 1

    # Handle styling.
    utils.make_table(
        ws,
        "designrequirements",
        min_row=1,
        max_row=rows,
        min_col=1,
        max_col=len(headers),
    )
    utils.apply_styling(ws, headers, defaults=OPTIONS)
    return ws, requirements


def add_behaviors_sheet(
    wb: Workbook, graph: Graph, components: List[Node]
) -> Tuple[Worksheet, List[Node]]:
    """Add a behavior requirements sheet to an Excel workbook.

    Arguments:
        wb: Excel workbook to add the behavior requirements sheet to.
        graph: Graph to fetch behavior nodes from.
        components: Component nodes to fetch behavior requirements for.

    Returns:
        Behavior requirements worksheet instance.
    """
    ws = wb.create_sheet("Behavior requirements")

    # Select requirements.
    requirements = utils.dedupe(
        [
            r
            for c in components
            for r in doc_utils.get_component_behaviors(c, graph, constraint=False)
        ]
        + [
            r
            for c in components
            for r in doc_utils.get_component_behaviors(c, graph, constraint=True)
        ]
    )

    # Handle headers.
    default_headers = [
        "instance path",
        "component definition",
        "form",
        "case",
        "when",
        "then",
        "comments",
    ]
    tags = utils.get_all_tags(requirements)
    headers = default_headers + tags
    ws.append(headers)
    rows = 1

    # Handle content.
    for requirement in requirements:
        info = requirement.annotations.esl_info
        tagged_content = [
            utils.format_multiline(info["tagged_comments"].get(tag, [])) for tag in tags
        ]
        if info["default"]:
            default_content = [
                node_path(requirement.name, italic=False, arrows=False, skip="world"),
                utils.parent_def(graph, requirement),
                info["form"],
                "default",
                "no other case applies",
                "\n".join(
                    text.designclause_text(clause["body"], graph, label=clause["name"])
                    for clause in info["default"]
                ),
                utils.format_multiline(info["comments"]),
            ]
            ws.append(default_content + tagged_content)
            rows += 1
        for case in info["cases"]:
            default_content = [
                node_path(requirement.name, italic=False, arrows=False, skip="world"),
                utils.parent_def(graph, requirement),
                info["form"],
                case["name"],
                "\n".join(
                    text.designclause_text(clause["body"], graph, label=clause["name"])
                    for clause in case["when_clauses"]
                ),
                "\n".join(
                    text.designclause_text(clause["body"], graph, label=clause["name"])
                    for clause in case["then_clauses"]
                ),
                utils.format_multiline(info["comments"]),
            ]
            ws.append(default_content + tagged_content)
            rows += 1

    # Handle styling.
    utils.make_table(
        ws,
        "behaviors",
        min_row=1,
        max_row=rows,
        min_col=1,
        max_col=len(headers),
    )
    utils.apply_styling(ws, headers, defaults=OPTIONS)
    return ws, requirements


def add_needs_sheet(
    wb: Workbook, graph: Graph, components: List[Node]
) -> Tuple[Worksheet, List[Node]]:
    """Add a needs sheet to an Excel workbook.

    Arguments:
        wb: Excel workbook to add the needs sheet to.
        graph: Graph to fetch need nodes from.
        components: Component nodes to fetch needs for.

    Returns:
        Needs worksheet instance.
    """
    ws = wb.create_sheet("Needs")

    # Select requirements.
    requirements = utils.dedupe(
        r for c in components for r in doc_utils.get_component_needs(c, graph)
    )

    # Handle headers.
    default_headers = [
        "instance path",
        "component definition",
        "subject",
        "text",
    ]
    tags = utils.get_all_tags(requirements)
    headers = default_headers + tags
    ws.append(headers)

    # Handle content.
    for requirement in requirements:
        info = requirement.annotations.esl_info
        default_content = [
            requirement.name.split(".")[-1],
            utils.parent_def(graph, requirement),
            var_path(graph[info["subject"]], italic=False, arrows=False, skip="world"),
            info["text"],
            utils.format_multiline(info["comments"]),
        ]
        tagged_content = [
            utils.format_multiline(info["tagged_comments"].get(tag, [])) for tag in tags
        ]
        ws.append(default_content + tagged_content)

    # Handle styling.
    utils.make_table(
        ws,
        "needs",
        min_row=1,
        max_row=1 + len(requirements),
        min_col=1,
        max_col=len(headers),
    )
    utils.apply_styling(ws, headers, defaults=OPTIONS)
    return ws, requirements


def add_variable_sheet(
    wb: Workbook, graph: Graph, components: List[Node]
) -> Tuple[Worksheet, List[Node]]:
    """Add a variable sheet to an Excel workbook.

    Arguments:
        wb: Excel workbook to add the needs sheet to.
        graph: Graph to fetch need nodes from.
        components: Component nodes to fetch needs for.

    Returns:
        Needs worksheet instance.
    """
    ws = wb.create_sheet("Variables")

    # Variables
    vrs = [n for n in graph.nodes if n.kind == "variable"]
    vrs = sorted(vrs, key=lambda x: x.name)

    # Handle headers.
    default_headers = ["instance path", "type", "domain", "units", "clarifaction"]
    tags = utils.get_all_tags(vrs)
    headers = default_headers + tags
    ws.append(headers)

    # Handle content.
    for var in vrs:
        info = var.annotations.esl_info
        default_content = list(get_var_table_row_elements(graph, var))
        tagged_content = [
            utils.format_multiline(info["tagged_comments"].get(tag, [])) for tag in tags
        ]
        ws.append(default_content + tagged_content)

    # Handle styling.
    utils.make_table(
        ws,
        "variables",
        min_row=1,
        max_row=1 + len(vrs),
        min_col=1,
        max_col=len(headers),
    )
    utils.apply_styling(ws, headers, defaults=OPTIONS)
    return ws, vrs


def add_overview_sheet(
    wb: Workbook,
    graph: Graph,
    components: List[Node],
    goals: List[Node],
    transformations: List[Node],
    designs: List[Node],
    behaviors: List[Node],
    needs: List[Node],
) -> Worksheet:
    """Add an overview sheet to an Excel workbook.

    Arguments:
        wb: Excel workbook to add the needs sheet to.
        components: Component nodes.
        goals: Goal requirement nodes.
        transformations: Transformation requirement nodes.
        designs: Design requirement nodes.
        behaviors: Behavior requirement nodes.
        needs: Need nodes.

    Returns:
        Overview worksheet instance.
    """
    ws = wb.create_sheet("Overview", index=0)

    # Select requirements.
    requirements = goals + transformations + designs + behaviors + needs

    # Handle headers.
    default_headers = [
        "instance name",
        "specification text",
        "component path",
        "component definition",
        "kind",
        "form",
        "comments",
    ]
    tags = utils.get_all_tags(components + requirements)
    headers = default_headers + tags
    ws.append(headers)

    # Handle content.
    for requirement in requirements:
        info = requirement.annotations.esl_info
        skip = utils.parent_component(requirement, skip=None)
        default_content = [
            requirement.name.split(".")[-1],
            text.requirement_text(requirement, graph, skip=skip),
            utils.parent_component(requirement, skip="world"),
            utils.parent_def(graph, requirement),
            utils.requirement_kind(requirement),
            None if requirement.kind == "need" else info["form"],
            utils.format_multiline(info["comments"]),
        ]
        tagged_content = [
            utils.format_multiline(info["tagged_comments"].get(tag, [])) for tag in tags
        ]
        ws.append(default_content + tagged_content)

    utils.make_table(
        ws,
        "overview",
        min_row=1,
        max_row=1 + len(requirements),
        min_col=1,
        max_col=len(headers),
    )
    utils.apply_styling(ws, headers, defaults=OPTIONS)

    return ws


def add_component_active_goals_sheet(
    wb: Workbook,
    graph: Graph,
    component: Node,
    flow_labels: Optional[List[str]] = None,
):
    ws_name = "Outgoing"
    ws = wb.create_sheet(ws_name)

    top_headers = ["instance path", "instance name"]
    ws.append(top_headers)
    split = component.name.rfind(".")
    path = component.name[:split]
    instance = component.name[split + 1 :]
    ws.append([path, instance])
    ws.append([])

    utils.make_table(
        ws,
        "outgoing-component-path",
        min_row=1,
        max_row=2,
        min_col=1,
        max_col=len(top_headers),
    )
    utils.apply_styling(ws, top_headers, defaults=OPTIONS, start_row=0, end_row=2)

    flows = set(flow_labels) if flow_labels else None

    functional_dependencies = utils.yield_functional_dependencies(graph, component, flows)

    function_specifications = set(
        [f for f in utils.yield_all_function_specifications(graph, functional_dependencies)]
    )

    active_goals = set(
        [(e, g) for (e, g) in utils.yield_active_functions(function_specifications, component.name)]
    )

    goalsdata = [
        data
        for (e, goal) in active_goals
        for data in utils.build_active_goal_data(graph, e, goal, function_specifications)
    ]

    headers = [
        "label",
        "target",
        "flows",
        "types",
        "subclause",
        "subject",
        "comparison",
        "bound",
        "unit",
    ]
    ws.append(headers)

    for data in goalsdata:
        ws.append([data.get(h, "") for h in headers])

    utils.make_table(
        ws,
        ws_name,
        min_row=4,
        max_row=4 + len(goalsdata),
        min_col=1,
        max_col=len(headers),
    )
    utils.apply_styling(ws, headers, defaults=OPTIONS, start_row=2, end_row=None)

    return ws


def add_component_passive_goals_sheet(
    wb: Workbook,
    graph: Graph,
    component: Node,
    flow_labels: Optional[List[str]] = None,
):
    ws_name = "Incoming"
    ws = wb.create_sheet(ws_name)

    top_headers = ["instance path", "instance name"]
    ws.append(top_headers)
    split = component.name.rfind(".")
    path = component.name[:split]
    instance = component.name[split + 1 :]
    ws.append([path, instance])
    ws.append([])

    utils.make_table(
        ws,
        "incoming-component-path",
        min_row=1,
        max_row=2,
        min_col=1,
        max_col=len(top_headers),
    )
    utils.apply_styling(ws, top_headers, defaults=OPTIONS, start_row=0, end_row=2)

    flows = set(flow_labels) if flow_labels else None

    functional_dependencies = utils.yield_functional_dependencies(graph, component, flows)

    function_specifications = utils.yield_all_function_specifications(
        graph, functional_dependencies
    )

    passive_goals = set(
        [
            (e, g)
            for (e, g) in utils.yield_passive_functions(function_specifications, component.name)
        ]
    )

    goalsdata = [
        data
        for (e, goal) in passive_goals
        for data in utils.build_passive_goal_data(graph, e, goal, function_specifications)
    ]

    headers = [
        "label",
        "source",
        "flows",
        "types",
        "subclause",
        "subject",
        "comparison",
        "bound",
        "unit",
    ]
    ws.append(headers)

    for data in goalsdata:
        ws.append([data.get(h, "") for h in headers])

    utils.make_table(
        ws,
        ws_name,
        min_row=4,
        max_row=4 + len(goalsdata),
        min_col=1,
        max_col=len(headers),
    )
    utils.apply_styling(ws, headers, defaults=OPTIONS, start_row=2, end_row=None)

    return ws



from pathlib import Path
from shutil import rmtree

import pytest
from pytest_examples import CodeExample, EvalExample, find_examples
from ragraph.graph import Graph

DOCS_DIR = Path(__file__).parent.parent / "docs"
DOCS_FILES = DOCS_DIR.glob("**/*.md")


@pytest.mark.parametrize("example", find_examples(*DOCS_FILES), ids=str)
def test_docs(
    example: CodeExample,
    eval_example: EvalExample,
    datadir: Path,
    docs_gendir: Path,
    pump_example_graph: Graph,
    tmpdir,
):
    """Test the package's documentation."""

    if example.prefix_settings().get("skip"):
        return

    pump_esl = datadir / "specs" / "pump_example.esl"
    globals = dict(
        pump_esl=pump_esl,
        path_to_esl_file=pump_esl,
        path_to_esl_dir=datadir / "esl",
        path_to_extra_esl_file=datadir / "esl" / "noodstopketen" / "level_0" / "preamble.esl",
        pump_prologue=datadir / "doc" / "md" / "prologue.md",
        pump_epilogue=datadir / "doc" / "md" / "epilogue.md",
        generated=docs_gendir,
        pump_example_graph=pump_example_graph,
        graph=pump_example_graph,
        cfv_path=docs_gendir / "cfv_mdm.svg",
        hierarchy_path=docs_gendir / "hierarchy_diagram",
        functional_dependency_path=docs_gendir / "functional_dependency_diagram",
        functional_context_path=docs_gendir / "functional_context_diagram",
        function_chain_path=docs_gendir / "function_chain_diagram",
        functional_traceability_path=docs_gendir / "functional_traceability_diagram",
    )

    eval_example.set_config(
        line_length=88,
        target_version="py311",
        ruff_select=["E", "W", "I"],
        ruff_ignore=["F821"],
    )
    if eval_example.update_examples:
        eval_example.format_ruff(example)
        eval_example.run_print_update(example, module_globals=globals)
    else:
        eval_example.lint_ruff(example)
        eval_example.run_print_check(example, module_globals=globals)

    # Temporary file cleanup because of rich document generation.
    rmtree(Path(__file__).parent.parent / "images", ignore_errors=True)



from pathlib import Path

from ragraph.io.canopy import to_canopy

from raesl import datasets
from raesl.canopy import add_canopy_annotations


def test_canopy(tmpdir, gendir, inspect):
    graph = datasets.get("pump")
    add_canopy_annotations(graph)

    if inspect:
        outfile = Path(gendir) / "pump-canopy-test.json"
    else:
        outfile = Path(tmpdir) / "pump-canopy-test.json"

    to_canopy(graph, outfile)

    graph = datasets.get("rally-car")
    add_canopy_annotations(graph)

    if inspect:
        outfile = Path(gendir) / "rally-car-test.json"
    else:
        outfile = Path(tmpdir) / "rally-car-test.json"

    to_canopy(graph, outfile)
